#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mem.h>

#define FALSE 0
#define TRUE -1


int main(int argc, char * argv[])
{
   FILE * in;
   long w, h, version, act, tag_type, f, i, n, p;
   int  c, x, y, nblayer, i2, n2;
   long f_num, s_num, t_num, w_num, obj_type, obj_id;
/*
        type1_lookup[5][60],
        type2_lookup[5][150]
*/
   int  lay_stream[12],
        orientation_lookup[25] = {
           0x00, 0x01, 0x02, 0x01, 0x02, 0x03, 0x03, 0x05, 0x05, 0x06, 0x06,
           0x07, 0x07, 0x08, 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10,
           0x11, 0x12, 0x14
        };
   char lay_name[13][15] = {
        {"NULL"},
        {"WALL 1"},{"WALL 2"},{"WALL 3"},{"WALL 4"},
        {"ORIENTATION 1"}, {"ORIENTATION 2"}, {"ORIENTATION 3"}, {"ORIENTATION 4"},
        {"FLOOR 1"}, {"FLOOR 2"},
        {"SHADOW"},
        {"TAG"}
   };


   if (argc == 1)
   {
      printf("Syntaxe : ds1info <file.ds1>\n");
      exit(0);
   }
   in = fopen(argv[1], "rb");
   if (in == NULL)
   {
      printf("can't open %s\n", argv[1]);
      exit(1);
   }
   printf("=======================\n%s\n=======================\n", argv[1]);

   w_num = 0; // # of wall & direction layers
   f_num = 0; // # of floor layer
   s_num = 1; // # of shadow layer, always here
   t_num = 0; // # of tag layer
      
   // version
   fread(&version, 1, 4, in);
   printf("version  = %li\n", version);
   
   // width
   fread(&w, 1, 4, in);
   w++;
   printf("width    = %li\n", w);

   // height
   fread(&h, 1, 4, in);
   h++;
   printf("height   = %li\n", h);

   // act
   act = 0;
   if (version >= 8)
   {
      fread(&act, 1, 4, in);
      if (act > 4)
         act = 4;
      printf("act      = %li\n", act + 1);
   }
   else
      printf("no act (version < 8)\n");

   // tag_type
   tag_type = 0;
   if (version >= 10)
   {
      fread(&tag_type, 1, 4, in);
      printf("tag_type = %li (tag layer, ", tag_type);
      if (tag_type)
         printf("type %li)\n", tag_type);
      else
         printf("not present)\n");

      // adjust eventually the # of groups tags layer
      if ((tag_type == 1) || (tag_type == 2))
         t_num = 1;
   }
   else
      printf("no tag_type (version < 10)\n");

   // filenames
   if (version >= 3)
   {
      fread(&f, 1, 4, in);
      printf("# files = %li\n", f);
      for (i=0; i<f; i++)
      {
         printf("   file %li = ", i + 1);
         c = fgetc(in);
         while (c != 0)
         {
            printf("%c", c > 32 ? c : 32);
            c = fgetc(in);
         }
         printf("\n");
      }
   }
   else
      printf("no files (version < 3)\n");

   // skip 2 dwords ?
   if ((version >= 9) && (version <= 13))
      fseek(in, 8, SEEK_CUR);

   // number of wall & floor layers
   if (version >= 4)
   {
      fread(&w_num, 1, 4, in);
      printf("w_num   = %li (wall layers)\n", w_num);

      // number of floor layers
      if (version >= 16)
      {
         fread(&f_num, 1, 4, in);
         printf("f_num   = %li (floor layers)\n", f_num);
      }
      else
      {
         f_num = 1; // default # of floor layer
         printf("no f_num (version < 16)\n");
      }
   }
   else // in version < 4
   {
      // # of layers hardcoded
      w_num = 1;
      f_num = 1;
      t_num = 1;
      printf("no w_num & no f_num (version < 4)\n");
   }

   // search the number of layers

   printf("\n# layers = (2 * %li walls) + %li floors + %li shadow "
          "+ %li tag\n",
          w_num, f_num, s_num, t_num
   );

   // which order ?
   if (version < 4)
   {
      lay_stream[0] =  1; // wall 1
      lay_stream[1] =  9; // floor 1
      lay_stream[2] =  5; // direction 1
      lay_stream[3] = 12; // tag
      lay_stream[4] = 11; // shadow
      i = 5;
   }
   else
   {
      i=0;
      for (x=0; x<w_num; x++)
      {
         lay_stream[i++] = 1 + x; // wall x
         lay_stream[i++] = 5 + x; // direction x
      }
      for (x=0; x<f_num; x++)
         lay_stream[i++] = 9 + x; // floor x
      if (s_num)
         lay_stream[i++] = 11;    // shadow
      if (t_num)
         lay_stream[i++] = 12;    // tag
   }
   nblayer = i;
   printf("         = %li layers\n", nblayer);
      
   // tiles of layers
   for (i=0; i<nblayer; i++)
   {
      printf("\nLayer %s (offset %08lX)\n",
         lay_name[lay_stream[i]],
         ftell(in)
      );
      for (y=0; y<h; y++)
      {
         for (x=0; x<w; x++)
         {
            // read 1 tile
            if (fread(&n, 1, 4, in) < 4)
            {
               printf("\n\nEOF\n");
               fclose(in);
               return;
            }
            else
            {
               if (version < 7)
               {
                  if ((lay_stream[i] >= 5) && (lay_stream[i] <= 8))
                  {
                     // convert direction layer
                     n = orientation_lookup[n];
                  }
               }
               printf(" %08lX", n);
            }
         }
         printf("\n");
      }
      printf("\n");
   }
   

   // objects
   if (version < 2)
      printf("no objects (version < 2)\n");
   else
   {
      printf("%08lX : # objects = ", ftell(in));
      fread(&n, 4, 1, in);
      printf("%li\n", n);
      if (n)
      {
         printf("   Num     Type       Id        X        Y    Flags\n");
         printf("   --- -------- -------- -------- -------- --------\n");
      }
      for (i=0; i<n; i++)
      {
         printf("   %3i", i);

         fread(&obj_type, 1, 4, in); // type
         fread(&obj_id,   1, 4, in); // id

/*
         switch (obj_type)
         {
            case 1 :
               if (version >= 5)
               {
                  obj_id = type1_lookup[act][obj_id];
                  if (act == 2) // act 3
                  {
                     if (obj_id == 297)
                     {
                        obj_id = 382;
                        obj_type = 2;
                     }
                     else if (obj_id == 366)
                     {
                        obj_id = 404;
                        obj_type = 2;
                     }
                  }
                  else if (act == 4) // act 5
                  {
                     if ((obj_id >= 537) && (obj_id <= 539))
                     {
                        obj_id = 1013 - obj_id;
                        obj_type = 2;
                     }
                     else if (obj_id == 514)
                     {
                        obj_id = 461;
                        obj_type = 2;
                     }
                  }
               }

               if (obj_id == 575)
               {
                  // set bad id
                  obj_id = -1;
               }
               break;

            case 2 :
               if ((obj_id == 573) || ((version > 5) && (obj_id >= 150)))
               {
                  // set bad id
                  obj_id = -1;
               }
               else
                  obj_id = type2_lookup[act][obj_id];
               break;

            case 4 :
               // I don't handle them
               break;

            default :
               break;
         }
*/

         // end_read_object

         printf(" %8li", obj_type);
         printf(" %8li", obj_id);

         fread(&p, 1, 4, in); // x
         printf(" %8li", p);
         fread(&p, 1, 4, in); // y
         printf(" %8li", p);

         if (version > 5)
         {
            fread(&p, 1, 4, in);  // flags
            printf(" %8li", p);
         }
         printf("\n");
      }
      printf("\n");
   }
   
   // groups for the tag layer
   if ((version >= 12) && ((tag_type == 1) || (tag_type == 2)))
   {
      // skip 1 dword ?
      if (version >= 18)
      {
         fread(&n, 4, 1, in);
         printf("skiping 1 dword : %08li\n\n", n);
      }
         
      printf("%08lX : GROUPS = ", ftell(in));
      fread(&n, 4, 1, in);
      printf("%li\n", n);
      if (n)
      {
         printf("   Num   Tile X   Tile Y    Width   Height         ?\n");
         printf("   --- -------- -------- -------- --------  --------\n");
      }
      for (i=0; i<n; i++)
      {
         printf("   %3i", i + 1);

         if (fread(&p, 1, 4, in) == 4)
            printf(" %8li", p);
         if (fread(&p, 1, 4, in) == 4)
            printf(" %8li", p);
         if (fread(&p, 1, 4, in) == 4)
            printf(" %8li", p);
         if (fread(&p, 1, 4, in) == 4)
            printf(" %8li", p);

         if (version >= 13)
         {
            if (fread(&p, 1, 4, in) == 4)
               printf(" %8li", p);
         }
         printf("\n");
      }
      printf("\n");
   }

   // npc paths
   if (version >= 14)
   {
      c = fread(&n, 1, 4, in);
      if (c == 4)
      {
         printf("%08lX : %li NPC paths\n", ftell(in), n);
         for (i=0; i<n; i++)
         {
            fread(&n2, 1, 4, in);
            printf("\nnpc %i : %li paths, object at coord (", i, n2);

            fread(&p, 1, 4, in);
            printf("%li, ", p);
            fread(&p, 1, 4, in);
            printf("%li) [%08lX]\n", p, ftell(in));

            printf("   Num        X        Y   Action\n");
            printf("   --- -------- -------- --------\n");
            for (i2=0; i2<n2; i2++)
            {
               printf("   %3i", i2);

               fread(&p, 1, 4, in);
               printf(" %8li", p);
               fread(&p, 1, 4, in);
               printf(" %8li", p);
               if (version >= 15)
               {
                  fread(&p, 1, 4, in);
                  printf(" %8li\n", p);
               }
               else
                  printf(" %8li\n", 1);
            }
        }
      }
      else
         printf("no NPC paths at all\n");
   }


   // until eof (just in case of out of sync)
   c = fread(&n, 1, 4, in);
   if (c == 4)
   {
      printf("\n%08lX : unknown datas\n", ftell(in));
      while (c == 4)
      {
         printf("%08X\n", n);
         c = fread(&n, 1, 4, in);
      }
   }
   if ((c>0) && (c != 4))
      printf("\n(%08X)\n", n);
      
   fclose(in);
}

